html,
body {
  height: 100%;
  min-height: 100%;
  margin: 0;
  overflow: hidden;
}

#terminal-container {
  width: auto;
  height: 100%;
  margin: 0 auto;
  padding: 0;
  .terminal {
    padding: 5px;
    height: calc(100% - 10px);
  }
}
